@@ -22,12 +22,13 @@ __all__ = [ 'adm', 'backends', 'block_cache', 'common', 'calc_mro',
             'REV_VER_MAP', 'RELEASE', 'BUFSIZE',
             'CTRL_NAME', 'CTRL_INODE' ]
 
-VERSION = '2.15'
+VERSION = '2.22'
 RELEASE = '%s' % VERSION
 
-# TODO: On next upgrade, remove pickle support from
-# common.py:load_params().
-CURRENT_FS_REV = 22
+# TODO: On next revision bump, consider removing support for TIME
+# values from deltadump.pyx. It is only present to allow upgrades
+# from revisions <= 22.
+CURRENT_FS_REV = 23
 
 # Buffer size when writing objects
 BUFSIZE = 64 * 1024
@@ -39,6 +40,7 @@ CTRL_INODE = 2
 # Maps file system revisions to the last S3QL version that
 # supported this revision.
 REV_VER_MAP = {
+    22: '2.16',
     21: '2.13',
     20: '2.9',
     16: '1.15',
