@@ -31,6 +31,7 @@ import hmac
 import re
 import tempfile
 import time
+import ssl
 import urllib.parse
 
 C_DAY_NAMES = [ 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun' ]
@@ -126,6 +127,15 @@ class Backend(AbstractBackend, metaclass=ABCDocstMeta):
         conn.timeout = int(self.options.get('tcp-timeout', 20))
         return conn
 
+    @staticmethod
+    def _tag_xmlns_uri(elem):
+        '''Extract the XML namespace (xmlns) URI from an element'''
+        if elem.tag[0] == '{':
+            uri, ignore, tag = elem.tag[1:].partition("}")
+        else:
+            uri = None
+        return uri
+
     # This method is also used implicitly for the retry handling of
     # `gs.Backend._get_access_token`. When modifying this method, do not forget
     # to check if this makes it unsuitable for use by `_get_access_token` (in
@@ -133,7 +143,8 @@ class Backend(AbstractBackend, metaclass=ABCDocstMeta):
     @copy_ancestor_docstring
     def is_temp_failure(self, exc): #IGNORE:W0613
         if isinstance(exc, (InternalError, BadDigestError, IncompleteBodyError,
-                            RequestTimeoutError, OperationAbortedError, SlowDownError)):
+                            RequestTimeoutError, OperationAbortedError,
+                            SlowDownError, ServiceUnavailableError)):
             return True
 
         elif is_temp_network_error(exc):
@@ -150,6 +161,12 @@ class Backend(AbstractBackend, metaclass=ABCDocstMeta):
                or exc.status == 408)):
             return True
 
+        # Temporary workaround for https://bitbucket.org/nikratio/s3ql/issues/87.
+        # We still need to find a proper string
+        elif (isinstance(exc, ssl.SSLError) and
+              str(exc).startswith('[SSL: BAD_WRITE_RETRY]')):
+            return True
+
         return False
 
     # NOTE: ! This function is also used by the swift backend. !
@@ -215,7 +232,6 @@ class Backend(AbstractBackend, metaclass=ABCDocstMeta):
         keys_remaining = True
         marker = self.prefix + start_after
         prefix = self.prefix + prefix
-        ns_p = self.xml_ns_prefix
 
         while keys_remaining:
             log.debug('requesting with marker=%s', marker)
@@ -233,15 +249,26 @@ class Backend(AbstractBackend, metaclass=ABCDocstMeta):
                 itree = iter(ElementTree.iterparse(self.conn, events=("start", "end")))
                 (event, root) = next(itree)
 
+                root_xmlns_uri = self._tag_xmlns_uri(root)
+                if root_xmlns_uri is None:
+                    root_xmlns_prefix = ''
+                else:
+                    # Validate the XML namespace
+                    root_xmlns_prefix = '{%s}' % (root_xmlns_uri, )
+                    if root_xmlns_prefix != self.xml_ns_prefix:
+                        log.error('Unexpected server reply to list operation:\n%s',
+                                  self._dump_response(resp, body=None))
+                        raise RuntimeError('List response has %s as root tag, unknown namespace' % root.tag)
+
                 for (event, el) in itree:
                     if event != 'end':
                         continue
 
-                    if el.tag == ns_p + 'IsTruncated':
+                    if el.tag == root_xmlns_prefix + 'IsTruncated':
                         keys_remaining = (el.text == 'true')
 
-                    elif el.tag == ns_p + 'Contents':
-                        marker = el.findtext(ns_p + 'Key')
+                    elif el.tag == root_xmlns_prefix + 'Contents':
+                        marker = el.findtext(root_xmlns_prefix + 'Key')
                         yield marker[len(self.prefix):]
                         root.clear()
 
@@ -404,9 +431,12 @@ class Backend(AbstractBackend, metaclass=ABCDocstMeta):
             return
         body = self.conn.readall()
         root = self._parse_xml_response(resp, body)
-        if root.tag == self.xml_ns_prefix + 'CopyObjectResult':
+
+        # Some S3 implemenentations do not have a namespace on
+        # CopyObjectResult.
+        if root.tag in [self.xml_ns_prefix + 'CopyObjectResult', 'CopyObjectResult']:
             return
-        elif root.tag == 'Error':
+        elif root.tag in [self.xml_ns_prefix + 'Error', 'Error']:
             raise get_S3Error(root.findtext('Code'), root.findtext('Message'),
                               resp.headers)
         else:
@@ -590,7 +620,7 @@ class Backend(AbstractBackend, metaclass=ABCDocstMeta):
     def __str__(self):
         return 's3c://%s/%s/%s' % (self.hostname, self.bucket_name, self.prefix)
 
-    def _authorize_request(self, method, path, headers, subres):
+    def _authorize_request(self, method, path, headers, subres, query_string):
         '''Add authorization information to *headers*'''
 
         # See http://docs.amazonwebservices.com/AmazonS3/latest/dev/RESTAuthentication.html
@@ -616,7 +646,9 @@ class Backend(AbstractBackend, metaclass=ABCDocstMeta):
             auth_strs.append('%s:%s\n' % (hdr, val))
 
         # Always include bucket name in path for signing
-        sign_path = urllib.parse.quote('/%s%s' % (self.bucket_name, path))
+        if self.hostname.startswith(self.bucket_name):
+            path = '/%s%s' % (self.bucket_name, path)
+        sign_path = urllib.parse.quote(path)
         auth_strs.append(sign_path)
         if subres:
             auth_strs.append('?%s' % subres)
@@ -638,11 +670,12 @@ class Backend(AbstractBackend, metaclass=ABCDocstMeta):
         if not isinstance(headers, CaseInsensitiveDict):
             headers = CaseInsensitiveDict(headers)
 
-        self._authorize_request(method, path, headers, subres)
-
-        # Construct full path
         if not self.hostname.startswith(self.bucket_name):
             path = '/%s%s' % (self.bucket_name, path)
+        headers['host'] = self.hostname
+
+        self._authorize_request(method, path, headers, subres, query_string)
+
         path = urllib.parse.quote(path)
         if query_string:
             s = urllib.parse.urlencode(query_string, doseq=True)
@@ -1002,5 +1035,6 @@ class SignatureDoesNotMatchError(S3Error, AuthenticationError): pass
 class OperationAbortedError(S3Error): pass
 class RequestTimeoutError(S3Error): pass
 class SlowDownError(S3Error): pass
+class ServiceUnavailableError(S3Error): pass
 class RequestTimeTooSkewedError(S3Error): pass
 class NoSuchBucketError(S3Error, DanglingStorageURLError): pass
